@@ -1,6 +1,6 @@
-;;; projectile-test.el
+;;; projectile-test.el --- Projectile's test suite -*- lexical-binding: t -*-
 
-;; Copyright © 2011-2020 Bozhidar Batsov
+;; Copyright © 2011-2021 Bozhidar Batsov
 
 ;; Author: Bozhidar Batsov <bozhidar@batsov.com>
 
@@ -21,24 +21,25 @@
 
 ;;; Commentary:
 
-;; This file is part of Projectile.
+;; Projectile's Buttercup-powered test suite.  You can run it either
+;; interactively or via the command-line (e.g. via eldev test).
 
 ;;; Code:
 
 (require 'projectile)
 (require 'buttercup)
 
-
+;; Useful debug information
 (message "Running tests on Emacs %s" emacs-version)
 
 ;; TODO: Revise this init logic
-(let* ((current-file (if load-in-progress load-file-name (buffer-file-name)))
-       (source-directory (locate-dominating-file current-file "Cask"))
-       ;; Do not load outdated byte code for tests
-       (load-prefer-newer t))
-  ;; Load the file under test
-  (load (expand-file-name "projectile" source-directory))
-  (setq projectile-test-path (expand-file-name "test" source-directory)))
+(defvar projectile-test-path (let* ((current-file (if load-in-progress load-file-name (buffer-file-name)))
+                                    (source-directory (locate-dominating-file current-file "Eldev"))
+                                    ;; Do not load outdated byte code for tests
+                                    (load-prefer-newer t))
+                               ;; Load the file under test
+                               (load (expand-file-name "projectile" source-directory))
+                               (expand-file-name "test" source-directory)))
 
 ;;; Test Utilities
 (defmacro projectile-test-with-sandbox (&rest body)
@@ -87,12 +88,54 @@ You'd normally combine this with `projectile-test-with-sandbox'."
      (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
      ,@body))
 
+;;; To avoid a macro, we could write a single "it" and iterate over a list of
+;;; functions. However, it would require manually rescuing the error and manually
+;;; re-throwing the error to include the function name in the error string, to
+;;; make it clear which function we're dealing with.  That could be avoided if
+;;; buttercup allowed us to specify a custom error message like this:
+;;;
+;;; (expect (funcall 'foo) :to-throw 'error nil "Custom error message here")
+(defmacro assert-friendly-error-when-no-project (fn)
+  "Write a test that ensures FN throws a friendly error when called without a project."
+  (let ((description (concat "when calling " (symbol-name fn) " without a project")))
+    `(describe
+      ,description
+      :var ((fn ',fn))
+      (it "throws a friendly error"
+          (projectile-test-with-sandbox
+           (projectile-test-with-files
+            ("index.html")
+            (find-file-noselect "index.html" t)
+            ;; Avoid "the current buffer is not visiting a file" error
+            (write-file "index.html")
+            (spy-on 'projectile-project-root :and-return-value nil)
+            (let ((projectile-require-project-root t))
+              (expect (call-interactively fn)
+                      :to-throw
+                      'error
+                      (list (concat "Projectile cannot find a project definition in "
+                                    default-directory))))))))))
+
 (defun projectile-test-tmp-file-path ()
   "Return a filename suitable to save data to in the test temp directory."
   (concat projectile-test-path
           "/tmp/temporary-file-" (format "%d" (random))
           ".eld"))
 
+(defun file-handler-for-tests (operation &rest args)
+  "Handler for # files.
+Just delegates OPERATION and ARGS for all operations except for`shell-command`'."
+  (let ((inhibit-file-name-handlers
+         (cons 'file-handler-for-tests
+               (and (eq inhibit-file-name-operation operation)
+                  inhibit-file-name-handlers)))
+        (inhibit-file-name-operation operation))
+    (cond ((eq operation 'shell-command) (let ((default-directory ""))
+                                           (shell-command "echo magic" t)))
+          (t (apply operation args)))))
+
+(add-to-list 'file-name-handler-alist (cons "^#" 'file-handler-for-tests))
+
 ;;; Tests
 (describe "projectile-project-name"
   (it "return projectile-project-name when present"
@@ -119,6 +162,20 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (expect (projectile-expand-root "foo/bar") :to-equal "/path/to/project/foo/bar")
     (expect (projectile-expand-root "./foo/bar") :to-equal "/path/to/project/foo/bar")))
 
+(describe "projectile--combine-plists"
+ (it "Items in second plist override elements in first"
+   (expect (projectile--combine-plists
+            '(:foo "foo" :bar "bar")
+            '(:foo "foo" :bar "foo" :foobar "foobar"))
+           :to-equal
+           '(:foo "foo" :bar "foo" :foobar "foobar")))
+ (it "Nil elements in second plist do not override elements in first"
+   (expect (projectile--combine-plists
+            '(:foo "foo" :bar "bar")
+            '(:foo "foo" :bar nil :foobar "foobar"))
+           :to-equal
+           '(:foo "foo" :bar "bar" :foobar "foobar"))))
+
 (describe "projectile-register-project-type"
   (it "prepends new projects to projectile-project-types"
     (projectile-register-project-type 'foo '("Foo"))
@@ -126,11 +183,47 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (projectile-register-project-type 'bar '("Bar"))
     (expect (caar projectile-project-types) :to-equal 'bar)))
 
+(describe "projectile-update-project-type"
+  :var ((mock-projectile-project-types
+           '((foo marker-files "marker-file"
+                  project-file "project-file"
+                  compilation-dir "compilation-dir"
+                  configure-command "configure"
+                  compile-command "compile"
+                  test-command "test"
+                  install-command "install"
+                  package-command "package"
+                  run-command "run"))))
+  (it "Updates existing project in projectile-project-types"
+    (let ((projectile-project-types mock-projectile-project-types))
+      (projectile-update-project-type
+       'foo
+       :marker-files "marker-file2"
+       :test-suffix "suffix")
+      (expect projectile-project-types :to-equal
+              '((foo marker-files "marker-file2"
+                     project-file "project-file"
+                     compilation-dir "compilation-dir"
+                     configure-command "configure"
+                     compile-command "compile"
+                     test-command "test"
+                     install-command "install"
+                     package-command "package"
+                     run-command "run"
+                     test-suffix "suffix")))))
+  (it "Error when attempt to update nonexistant project type"
+    (let ((projectile-project-types mock-projectile-project-types))
+      (expect (projectile-update-project-type
+               'bar
+               :marker-files "marker-file"
+               :test-suffix "suffix")
+              :to-throw))))
+
 (describe "projectile-project-type"
   (it "detects the type of Projectile's project"
-    (expect (projectile-project-type) :to-equal 'emacs-cask))
+    (expect (projectile-project-type) :to-equal 'emacs-eldev))
   (it "caches the project type"
-    (expect (gethash (projectile-project-root) projectile-project-type-cache) :to-equal 'emacs-cask)))
+    (expect (gethash (projectile-project-root) projectile-project-type-cache) :to-equal 'emacs-eldev)))
 
 (describe "projectile-ignored-directory-p"
   (it "checks if directory should be ignored"
@@ -150,11 +243,11 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (spy-on 'projectile-project-root :and-return-value "/path/to/project")
     (spy-on 'projectile-project-name :and-return-value "project")
     (spy-on 'projectile-project-ignored-files :and-return-value '("foo.js" "bar.rb"))
-    (let ((files'("/path/to/project/TAGS"
-                  "/path/to/project/foo.js"
-                  "/path/to/project/bar.rb"
-                  "/path/to/project/file1.log"
-                  "/path/to/project/file2.log"))
+    (let ((files '("/path/to/project/TAGS"
+                   "/path/to/project/foo.js"
+                   "/path/to/project/bar.rb"
+                   "/path/to/project/file1.log"
+                   "/path/to/project/file2.log"))
           (projectile-ignored-files '("TAGS" "file\d+\\.log")))
       (expect (projectile-ignored-files) :not :to-equal files)
       (expect (projectile-ignored-files) :to-equal '("/path/to/project/TAGS"
@@ -251,10 +344,23 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (spy-on 'file-truename :and-call-fake (lambda (filename) filename))
     (spy-on 'insert-file-contents :and-call-fake
             (lambda (filename)
-              (save-excursion (insert "\n-exclude\n+include\nno-prefix\n left-wspace\nright-wspace\t\n"))))
+              (save-excursion (insert "\n-exclude\n+include\n#may-be-a-comment\nno-prefix\n left-wspace\nright-wspace\t\n"))))
     (expect (projectile-parse-dirconfig-file) :to-equal '(("include/")
-                                                          ("exclude" "no-prefix" "left-wspace" "right-wspace")
-                                                          nil))))
+                                                          ("exclude"
+							   "#may-be-a-comment"
+							   "no-prefix"
+							   "left-wspace"
+							   "right-wspace")
+                                                          nil))
+    ;; same test - but with comment lines enabled using prefix '#'
+    (let ((projectile-dirconfig-comment-prefix ?#))
+      (expect (projectile-parse-dirconfig-file) :to-equal '(("include/")
+							    ("exclude"
+							     "no-prefix"
+							     "left-wspace"
+							     "right-wspace")
+							    nil)))
+    ))
 
 (describe "projectile-get-project-directories"
   (it "gets the list of project directories"
@@ -292,18 +398,34 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (expect (string-empty-p (projectile-get-sub-projects-command 'none)) :to-be-truthy)))
 
 (describe "projectile-files-via-ext-command"
-  (it "returns nil when command is nil or empty"
-    (expect (projectile-files-via-ext-command "" "") :not :to-be-truthy)
-    (expect (projectile-files-via-ext-command "" nil) :not :to-be-truthy)))
+          (it "returns nil when command is nil or empty or fails"
+              (expect (projectile-files-via-ext-command "" "") :not :to-be-truthy)
+              (expect (projectile-files-via-ext-command "" nil) :not :to-be-truthy)
+              (expect (projectile-files-via-ext-command "" "echo Not a file name! > &2") :not :to-be-truthy)
+              (expect (projectile-files-via-ext-command "" "echo filename") :to-equal '("filename")))
+
+          (it "supports magic file handlers"
+              (expect (projectile-files-via-ext-command "#magic#" "echo filename") :to-equal '("magic"))))
 
 (describe "projectile-mode"
-  (it "sets up hook functions"
+  (before-each
     (spy-on 'projectile--cleanup-known-projects)
-    (spy-on 'projectile-discover-projects-in-search-path)
+    (spy-on 'projectile-discover-projects-in-search-path))
+  (it "sets up hook functions"
     (projectile-mode 1)
     (expect (memq 'projectile-find-file-hook-function find-file-hook) :to-be-truthy)
     (projectile-mode -1)
-    (expect (memq 'projectile-find-file-hook-function find-file-hook) :not :to-be-truthy)))
+    (expect (memq 'projectile-find-file-hook-function find-file-hook) :not :to-be-truthy))
+  (it "respects projectile-auto-discover setting"
+    (unwind-protect
+        (progn
+          (let ((projectile-auto-discover nil))
+            (projectile-mode 1)
+            (expect 'projectile-discover-projects-in-search-path :not :to-have-been-called))
+          (let ((projectile-auto-discover t))
+            (projectile-mode 1)
+            (expect 'projectile-discover-projects-in-search-path :to-have-been-called)))
+      (projectile-mode -1))))
 
 (describe "projectile-relevant-known-projects"
   (it "returns a list of known projects"
@@ -455,16 +577,16 @@ You'd normally combine this with `projectile-test-with-sandbox'."
       (let ((projectile-project-root-files-bottom-up '("somefile" ".projectile"))
             (projectile-project-root-files '("otherfile" "framework.conf" "requirements.txt"))
             (projectile-project-root-files-top-down-recurring '(".svn" ".foo"))
-            (projectile-project-root-files-functions '(projectile-root-bottom-up
-                                                       projectile-root-top-down
-                                                       projectile-root-top-down-recurring)))
+            (projectile-project-root-functions '(projectile-root-bottom-up
+                                                 projectile-root-top-down
+                                                 projectile-root-top-down-recurring)))
         (projectile-test-should-root-in "projectA" "projectA/requirements/a/b/c/d/e/f/g")
         (projectile-test-should-root-in "projectA" "projectA/src/framework/lib")
         (projectile-test-should-root-in "projectA" "projectA/src/html")
 
-        (setq projectile-project-root-files-functions '(projectile-root-top-down
-                                                        projectile-root-top-down-recurring
-                                                        projectile-root-bottom-up))
+        (setq projectile-project-root-functions '(projectile-root-top-down
+                                                  projectile-root-top-down-recurring
+                                                  projectile-root-bottom-up))
         (projectile-test-should-root-in "projectA/requirements/a/b/c"
                                         "projectA/requirements/a/b/c/d/e/f/g")
         (projectile-test-should-root-in "projectA/src/framework"
@@ -475,9 +597,9 @@ You'd normally combine this with `projectile-test-with-sandbox'."
       (let ((projectile-project-root-files-bottom-up '("somefile" ".projectile"))
             (projectile-project-root-files '("otherfile" "noframework.conf"))
             (projectile-project-root-files-top-down-recurring '(".svn" ".foo"))
-            (projectile-project-root-files-functions '(projectile-root-top-down-recurring
-                                                       projectile-root-bottom-up
-                                                       projectile-root-top-down)))
+            (projectile-project-root-functions '(projectile-root-top-down-recurring
+                                                 projectile-root-bottom-up
+                                                 projectile-root-top-down)))
         (projectile-test-should-root-in "projectA/src" "projectA/src/framework/lib")
         (projectile-test-should-root-in "projectA/src" "projectA/src/html")
         (projectile-test-should-root-in "projectA/" "projectA/build/framework/lib"))
@@ -485,18 +607,18 @@ You'd normally combine this with `projectile-test-with-sandbox'."
       (let ((projectile-project-root-files-bottom-up '("somefile" "override"))
             (projectile-project-root-files '("otherfile" "anotherfile"))
             (projectile-project-root-files-top-down-recurring '("someotherfile" "yetanotherfile"))
-            (projectile-project-root-files-functions '(projectile-root-bottom-up
-                                                       projectile-root-top-down
-                                                       projectile-root-top-down-recurring)))
+            (projectile-project-root-functions '(projectile-root-bottom-up
+                                                 projectile-root-top-down
+                                                 projectile-root-top-down-recurring)))
         (projectile-test-should-root-in default-directory "projectA/src/framework/lib")
         (projectile-test-should-root-in default-directory "projectA/src/html"))
 
       (let ((projectile-project-root-files-bottom-up '("somecoolfile"))
             (projectile-project-root-files nil)
             (projectile-project-root-files-top-down-recurring '(".svn"))
-            (projectile-project-root-files-functions '(projectile-root-bottom-up
-                                                       projectile-root-top-down
-                                                       projectile-root-top-down-recurring)))
+            (projectile-project-root-functions '(projectile-root-bottom-up
+                                                 projectile-root-top-down
+                                                 projectile-root-top-down-recurring)))
         (projectile-test-should-root-in "projectA/src/" "projectA/src/")
         (projectile-test-should-root-in "projectA/src/" "projectA/src/html"))))))
 
@@ -620,7 +742,7 @@ You'd normally combine this with `projectile-test-with-sandbox'."
                  0 ;; Cached 1st of January 1970.
                  projectile-projects-cache-time)
 
-        (spy-on 'projectile-project-root :and-call-fake (lambda () (file-truename default-directory)))
+        (spy-on 'projectile-acquire-root :and-call-fake (lambda () (file-truename default-directory)))
         (spy-on 'projectile-project-vcs :and-return-value 'none)
         ;; After listing all the files, the cache should have been updated.
         (projectile-current-project-files)
@@ -761,6 +883,54 @@ You'd normally combine this with `projectile-test-with-sandbox'."
     (let ((projectile-known-projects nil))
       (expect (projectile-switch-project) :to-throw))))
 
+(describe "projectile-switch-project-by-name"
+  (it "calls the switch project action with project-to-switch's dir-locals loaded"
+    (defvar switch-project-foo)
+    (let ((foo 'bar)
+          (switch-project-foo)
+          (safe-local-variable-values '((foo . baz)))
+          (projectile-switch-project-action (lambda () (setq switch-project-foo foo))))
+      (projectile-test-with-sandbox
+        (projectile-test-with-files
+            ("project/"
+             "project/.dir-locals.el"
+             "project/.projectile")
+          (append-to-file
+           "((nil . ((foo . baz))))" nil "project/.dir-locals.el")
+          (projectile-add-known-project (file-name-as-directory (expand-file-name "project")))
+          (projectile-switch-project-by-name (file-name-as-directory (expand-file-name "project")))
+
+          (expect switch-project-foo :to-be 'baz)))))
+
+  (it "runs hooks from the project root directory"
+      (defvar hook-dir)
+      (let ((projectile-switch-project-action
+             (lambda () (switch-to-buffer (find-file-noselect "file" t))))
+            (hook (lambda () (setq hook-dir default-directory))))
+        (add-hook 'projectile-after-switch-project-hook hook)
+        (projectile-test-with-sandbox
+         (projectile-test-with-files
+          ("project/"
+           "project/file")
+          (let ((project-dir (file-name-as-directory (expand-file-name "project"))))
+            (projectile-add-known-project project-dir)
+            (projectile-switch-project-by-name project-dir)
+            (remove-hook 'projectile-after-switch-project-hook hook)
+
+            (expect hook-dir :to-equal project-dir))))))
+
+  (it "ensures the buffer is switched immediately"
+      (let ((projectile-switch-project-action
+             (lambda () (switch-to-buffer (find-file-noselect "file" t)))))
+        (projectile-test-with-sandbox
+         (projectile-test-with-files
+          ("project/"
+           "project/file")
+          (projectile-add-known-project (file-name-as-directory (expand-file-name "project")))
+          (projectile-switch-project-by-name (file-name-as-directory (expand-file-name "project")))
+
+          (expect (current-buffer) :to-be (get-file-buffer "project/file")))))))
+
 (describe "projectile-ignored-buffer-p"
   (it "checks if buffer should be ignored"
     (let ((projectile-globally-ignored-buffers '("*nrepl messages*" "*something*")))
@@ -956,7 +1126,16 @@ You'd normally combine this with `projectile-test-with-sandbox'."
        "project/package.json")
       (let ((projectile-indexing-method 'native))
         (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
-        (expect (projectile-detect-project-type) :to-equal 'rails-rspec))))))
+        (expect (projectile-detect-project-type) :to-equal 'rails-rspec)))))
+  (it "detects project-type for elisp eldev projects"
+    (projectile-test-with-sandbox
+     (projectile-test-with-files
+      ("project/"
+       "project/Eldev"
+       "project/project.el")
+      (let ((projectile-indexing-method 'native))
+        (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
+        (expect (projectile-detect-project-type) :to-equal 'emacs-eldev))))))
 
 (describe "projectile-dirname-matching-count"
   (it "counts matching dirnames ascending file paths"
@@ -1033,7 +1212,37 @@ You'd normally combine this with `projectile-test-with-sandbox'."
         (expect (projectile-test-file-p "src/Foo.cpp") :to-equal nil)
         (expect (projectile--find-matching-test "src/Foo.cpp") :to-equal '("test/Foo.cpp"))
         (expect (projectile--find-matching-test "src/Foo.cpp") :to-equal '("test/Foo.cpp"))
-        (expect (projectile--find-matching-file "test/Foo.cpp") :to-equal '("src/Foo.cpp"))))))
+        (expect (projectile--find-matching-file "test/Foo.cpp") :to-equal '("src/Foo.cpp")))))
+
+  (it "defers to test-dir property when it's set to a function"
+    (projectile-test-with-sandbox
+     (projectile-test-with-files-using-custom-project
+          ("src/foo/Foo.cpp"
+           "src/bar/Foo.cpp"
+           "test/foo/FooTest.cpp")
+          (:test-dir
+           (lambda (file-path)
+             (projectile-complementary-dir file-path "src" "test"))
+           :test-suffix "Test")
+          (expect (projectile--find-matching-test
+                   (projectile-expand-root "src/bar/Foo.cpp"))
+                  :to-equal
+                  (list "test/bar/FooTest.cpp")))))
+
+  (it "defers to src-dir property when it's set to a function"
+    (projectile-test-with-sandbox
+     (projectile-test-with-files-using-custom-project
+          ("src/foo/Foo.cpp"
+           "src/bar/Foo.cpp"
+           "test/foo/FooTest.cpp")
+          (:src-dir
+           (lambda (file-path)
+             (projectile-complementary-dir file-path "test" "src"))
+           :test-suffix "Test")
+          (expect (projectile--find-matching-file
+                   (projectile-expand-root "test/foo/FooTest.cpp"))
+                  :to-equal
+                  (list "src/foo/Foo.cpp"))))))
 
 (describe "projectile--related-files"
   (it "returns related files for the given file"
@@ -1330,6 +1539,14 @@ You'd normally combine this with `projectile-test-with-sandbox'."
       (projectile-clear-known-projects)
       (expect projectile-known-projects :to-equal nil))))
 
+(describe "projectile-reset-known-projects"
+  (it "resets known projects"
+    (spy-on 'projectile-clear-known-projects)
+    (spy-on 'projectile-discover-projects-in-search-path)
+    (projectile-reset-known-projects)
+    (expect 'projectile-clear-known-projects :to-have-been-called)
+    (expect 'projectile-discover-projects-in-search-path :to-have-been-called)))
+
 (describe "projectile-test-ignored-directory-p"
   (it "ignores specified literal directory values"
     (spy-on 'projectile-ignored-directories :and-return-value '("/path/to/project/tmp"))
@@ -1414,3 +1631,240 @@ You'd normally combine this with `projectile-test-with-sandbox'."
 (describe "projectile-find-file-in-directory"
   (it "fails when called in a non-existing directory"
     (expect (projectile-find-file-in-directory "asdf") :to-throw)))
+
+(describe "projectile-dir-files-native"
+  (it "calculates ignored files and directories only once during recursion"
+    (projectile-test-with-sandbox
+     (projectile-test-with-files
+      ("projectA/"
+       "projectA/.svn/"
+       "projectA/src/.svn/"
+       "projectA/src/html/.svn/"
+       "projectA/.git/"
+       "projectA/src/html/"
+       "projectA/src/framework/lib/"
+       "projectA/src/framework.conf"
+       "projectA/src/html/index.html"
+       "projectA/.projectile")
+
+      ;; verify that indexing only invokes these funcs once during recursion
+      (spy-on 'projectile-ignored-files :and-call-through)
+      (spy-on 'projectile-ignored-directories :and-call-through)
+      (spy-on 'projectile-globally-ignored-directory-names :and-call-through)
+
+      (projectile-dir-files-native "projectA/")
+      (expect 'projectile-ignored-files :to-have-been-called-times 1)
+      (expect 'projectile-globally-ignored-directory-names :to-have-been-called-times 1)
+      (expect 'projectile-ignored-directories :to-have-been-called-times 1))))
+  (it "ignores globally ignored directories when using native indexing"
+      (projectile-test-with-sandbox
+       (projectile-test-with-files
+        ("project/"
+         "project/.ignoreme/"
+         "project/.ignoreme/should_ignore"
+         "project/src/"
+         "project/src/.ignoreme/"
+         "project/src/.ignoreme/should_ignore"
+         "project/config.conf")
+
+        (setq projectile-globally-ignored-directories '(".ignoreme"))
+        (expect (projectile-dir-files-native "project") :to-equal '("config.conf"))))))
+
+(describe "projectile-process-current-project-buffers-current"
+  (it "expects projectile-process-current-project-buffers and
+projectile-process-current-project-buffers-current to have similar behaviour"
+    (projectile-test-with-sandbox
+     (projectile-test-with-files
+      ("projectA/"
+       "projectA/.projectile"
+       "projectA/bufferA"
+       "projectA/fileA"
+       "projectA/dirA/"
+       "projectA/dirA/fileC")
+      (let ((list-a '())
+            (list-b '()))
+        (projectile-process-current-project-buffers (lambda (b) (push b list-a)))
+        (projectile-process-current-project-buffers-current (lambda () (push (current-buffer) list-b)))
+        (expect list-a :to-equal list-b))))))
+
+(describe "projectile--impl-name-for-test-name"
+  :var ((mock-projectile-project-types
+         '((foo test-suffix "Test")
+           (bar test-prefix "Test"))))
+  (it "removes suffix from test file"
+    (cl-letf (((symbol-function 'projectile-project-type) (lambda () 'foo))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-name-for-test-name "FooTest.cpp")
+              :to-equal
+              "Foo.cpp")))
+  (it "removes prefix from test file"
+    (cl-letf (((symbol-function 'projectile-project-type) (lambda () 'bar))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-name-for-test-name "TestFoo.cpp")
+              :to-equal
+              "Foo.cpp"))))
+
+(describe "projectile-find-implementation-or-test"
+  (it "error when test file does not exist and projectile-create-missing-test-files is nil"
+    (cl-letf (((symbol-function 'projectile-test-file-p) #'ignore)
+              ((symbol-function 'file-exists-p) #'ignore)
+              ((symbol-function 'projectile-expand-root) #'identity)
+              ((symbol-function 'projectile-find-matching-test) (lambda (file) "dir/foo"))
+              (projectile-create-missing-test-files nil))
+      (expect (projectile-find-implementation-or-test "foo") :to-throw))))
+
+(describe "projectile--impl-file-from-src-dir-fn"
+  :var ((mock-projectile-project-types
+         '((foo src-dir (lambda (impl-file) "/outer/foo/test/dir"))
+           (bar src-dir "not a function"))))
+  (it "returns result of projectile--complementary-file when src-dir property is a function"
+    (cl-letf (((symbol-function 'projectile--complementary-file)
+               (lambda (impl-file dir-fn file-fn) (funcall dir-fn impl-file)))
+              ((symbol-function 'projectile-project-type) (lambda () 'foo))
+              ((symbol-function 'projectile-project-root) (lambda () "foo"))
+              ((symbol-function 'file-relative-name) (lambda (f rel) f))
+              ((symbol-function 'file-exists-p) (lambda (file) t))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-file-from-src-dir-fn "foo") :to-equal "/outer/foo/test/dir")))
+  (it "returns file relative to project root"
+    (cl-letf (((symbol-function 'projectile--complementary-file)
+               (lambda (impl-file dir-fn file-fn) (funcall dir-fn impl-file)))
+              ((symbol-function 'projectile-project-type) (lambda () 'foo))
+              ((symbol-function 'projectile-project-root) (lambda () "/outer/foo"))
+              ((symbol-function 'file-exists-p) (lambda (file) t))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-file-from-src-dir-fn "/outer/foo/bar")
+              :to-equal
+              "test/dir")))
+  (it "returns nil when src-dir property is a not function"
+    (cl-letf (((symbol-function 'projectile-project-type) (lambda () 'bar))
+              ((symbol-function 'projectile-project-root) (lambda () "foo"))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-file-from-src-dir-fn "bar") :to-equal nil)))
+  (it "returns nil when src-dir function result is not an existing file"
+    (cl-letf (((symbol-function 'projectile--complementary-file)
+               (lambda (impl-file dir-fn file-fn) (funcall dir-fn impl-file)))
+              ((symbol-function 'projectile-project-type) (lambda () 'foo))
+              ((symbol-function 'projectile-project-root) (lambda () "/outer/foo"))
+              ((symbol-function 'file-exists-p) #'ignore)
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--impl-file-from-src-dir-fn "bar") :to-equal nil))))
+
+(describe "projectile--test-file-from-test-dir-fn"
+  :var ((mock-projectile-project-types
+         '((foo test-dir (lambda (impl-file) "/outer/foo/test/dir"))
+           (bar test-dir "not a function"))))
+  (it "returns result of projectile--complementary-file when test-dir property is a function"
+    (cl-letf (((symbol-function 'projectile--complementary-file)
+               (lambda (impl-file dir-fn file-fn) (funcall dir-fn impl-file)))
+              ((symbol-function 'projectile-project-type) (lambda () 'foo))
+              ((symbol-function 'projectile-project-root) (lambda () "foo"))
+              ((symbol-function 'file-relative-name) (lambda (f rel) f))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--test-file-from-test-dir-fn "foo") :to-equal "/outer/foo/test/dir")))
+  (it "returns file relative to project root"
+    (cl-letf (((symbol-function 'projectile-project-type) (lambda () 'foo))
+              ((symbol-function 'projectile-project-root) (lambda () "/outer/foo"))
+              ((symbol-function 'projectile--complementary-file)
+               (lambda (impl-file dir-fn file-fn) (funcall dir-fn impl-file)))
+              (projectile-project-types mock-projectile-project-types))
+      (expect (projectile--test-file-from-test-dir-fn "/outer/foo/bar")
+              :to-equal
+              "test/dir")))
+  (it "returns nil when test-dir property is a not function"
+    (cl-letf (((symbol-function 'projectile-project-type) (lambda () 'bar))
+              (projectile-project-types mock-projectile-project-types)
+              ((symbol-function 'projectile-project-root) (lambda () "foo")))
+      (expect (projectile--test-file-from-test-dir-fn "bar") :to-equal nil))))
+
+(describe "projectile--complementary-file"
+  (it "dir-fn and filename-fn applied correctly"
+    (cl-letf (((symbol-function 'file-exists-p) (lambda (file) t))
+              ((symbol-function 'dir-fn) (lambda (dir) "foo/test/dir"))
+              ((symbol-function 'filename-fn) (lambda (filename) "Foo.test")))
+      (expect (projectile--complementary-file
+               "foo/src/dir/Foo.impl"
+               #'dir-fn
+               #'filename-fn)
+              :to-equal "foo/test/dir/Foo.test"))))
+
+(describe "projectile--impl-to-test-dir"
+    :var ((mock-projectile-project-types
+           '((foo test-dir "test" src-dir "src")
+             (bar test-dir identity src-dir "src"))))
+    (it "replaces occurrences of src-dir with test-dir"
+      (cl-letf (((symbol-function 'projectile-project-root) (lambda () "foo"))
+                ((symbol-function 'projectile-project-type) (lambda () 'foo))
+                (projectile-project-types mock-projectile-project-types))
+        (expect (projectile--impl-to-test-dir "/foo/src/Foo") :to-equal "/foo/test/")))
+    (it "error signalled when test dir property is not a string"
+      (cl-letf (((symbol-function 'projectile-project-root) (lambda () "bar"))
+                ((symbol-function 'projectile-project-type) (lambda () 'bar))
+                (projectile-project-types mock-projectile-project-types))
+        (expect (projectile--impl-to-test-dir "/bar/src/bar") :to-throw))))
+
+(describe "projectile-run-shell-command-in-root"
+  (describe "when called directly in elisp"
+    (before-each (spy-on 'shell-command))
+    (describe "when called with all three paramters"
+      (it "expects to call shell-command with the same parameters"
+        (projectile-run-shell-command-in-root "cmd" "output-buffer" "error-buffer")
+        (expect 'shell-command :to-have-been-called-with "cmd" "output-buffer" "error-buffer")))
+    (describe "when called with only one optional paramter"
+      (it "expects to call shell-command with the same parameters"
+        (projectile-run-shell-command-in-root "cmd" "output-buffer")
+        (expect 'shell-command :to-have-been-called-with "cmd" "output-buffer" nil)))
+    (describe "when called with no optional paramters"
+      (it "expects to call shell-command with the same parameters"
+        (projectile-run-shell-command-in-root "cmd")
+        (expect 'shell-command :to-have-been-called-with "cmd" nil nil))))
+  (describe "when called interactively"
+    (before-each (spy-on 'shell-command))
+    (it "expects to be interactive"
+      (expect (interactive-form 'projectile-run-shell-command-in-root) :not :to-be nil))
+    (it "expects to call shell-command with the given command"
+      (funcall-interactively 'projectile-run-shell-command-in-root "cmd")
+      (expect 'shell-command :to-have-been-called-with "cmd" nil nil))))
+
+(describe "projectile-run-async-shell-command-in-root"
+  (describe "when called directly in elisp"
+    (before-each (spy-on 'async-shell-command))
+    (describe "when called with all three paramters"
+      (it "expects to call async-shell-command with the same parameters"
+        (projectile-run-async-shell-command-in-root "cmd" "output-buffer" "error-buffer")
+        (expect 'async-shell-command :to-have-been-called-with "cmd" "output-buffer" "error-buffer")))
+    (describe "when called with only one optional paramter"
+      (it "expects to call async-shell-command with the same parameters"
+        (projectile-run-async-shell-command-in-root "cmd" "output-buffer")
+        (expect 'async-shell-command :to-have-been-called-with "cmd" "output-buffer" nil)))
+    (describe "when called with no optional paramters"
+      (it "expects to call async-shell-command with the same parameters"
+        (projectile-run-async-shell-command-in-root "cmd")
+        (expect 'async-shell-command :to-have-been-called-with "cmd" nil nil))))
+  (describe "when called interactively"
+    (before-each (spy-on 'async-shell-command))
+    (it "expects to be interactive"
+      (expect (interactive-form 'projectile-run-async-shell-command-in-root) :not :to-be nil))
+    (it "expects to call async-shell-command with the given command"
+      (funcall-interactively 'projectile-run-async-shell-command-in-root "cmd")
+      (expect 'async-shell-command :to-have-been-called-with "cmd" nil nil))))
+
+;; A bunch of tests that make sure Projectile commands handle
+;; gracefully the case of being run outside of a project.
+(assert-friendly-error-when-no-project projectile-project-info)
+(assert-friendly-error-when-no-project projectile-display-buffer)
+(assert-friendly-error-when-no-project projectile-find-implementation-or-test-other-frame)
+(assert-friendly-error-when-no-project projectile-find-implementation-or-test-other-window)
+(assert-friendly-error-when-no-project projectile-find-other-file)
+(assert-friendly-error-when-no-project projectile-find-other-file-other-frame)
+(assert-friendly-error-when-no-project projectile-find-other-file-other-window)
+(assert-friendly-error-when-no-project projectile-find-test-file)
+(assert-friendly-error-when-no-project projectile-grep)
+(assert-friendly-error-when-no-project projectile-ibuffer)
+(assert-friendly-error-when-no-project projectile-project-buffers-other-buffer)
+(assert-friendly-error-when-no-project projectile-project-info)
+(assert-friendly-error-when-no-project projectile-regenerate-tags)
+(assert-friendly-error-when-no-project projectile-remove-current-project-from-known-projects)
+(assert-friendly-error-when-no-project projectile-switch-to-buffer)
+(assert-friendly-error-when-no-project projectile-switch-to-buffer-other-frame)
+(assert-friendly-error-when-no-project projectile-switch-to-buffer-other-window)
